package com.javacodegeeks.basics_of_servlets;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Objects;
import java.util.concurrent.ExecutorService;

import javax.servlet.AsyncContext;
import javax.servlet.ServletException;
import javax.servlet.annotation.WebInitParam;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.javacodegeeks.basics_of_servlets.listeners.SnailAsyncListener;

// only handles GET requests
// we initialize the configuration for this servlet with a WebInitParam representing the timeout in milliseconds for the asynchronous processing to take
// we also indicate to the container that this servlet supports asynchronous processing
@WebServlet(urlPatterns = "/snail/snailservlet", initParams = { @WebInitParam(name = "timeout", value = "5000") }, asyncSupported = true)
public class SnailServlet extends HttpServlet {

    private static final String TIMEOUT_PARAM_KEY = "timeout";

    private static final long serialVersionUID = 1L;

    protected final void doGet(final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {

        // required to put the request into asynchronous mode
        final AsyncContext asyncCtx = request.startAsync();

        // not needed, but we are interested in the events generated from the
        // asynchronous processing
        asyncCtx.addListener(new SnailAsyncListener());

        // we set our timeout for processing
        asyncCtx.setTimeout(Long.valueOf(getServletConfig().getInitParameter(TIMEOUT_PARAM_KEY)));

        // we retrieve our thread pool executor from the application wide
        // servlet context
        final ExecutorService executor = (ExecutorService) request.getServletContext().getAttribute(Constants.THREAD_POOL_EXECUTOR);

        // submit a runnable containing the AsyncContext for flusing the
        // response to
        executor.execute(new SnailHandler(asyncCtx));
    }

    private static final class SnailHandler implements Runnable {

        private AsyncContext asyncContext;

        // receives a handle to the AsyncContext in order to flush the response.
        public SnailHandler(final AsyncContext asyncCtx) {
            this.asyncContext = asyncCtx;
        }

        @Override
        public void run() {
            PrintWriter out = null;
            try {
                // our snail is on a Sunday cruise
                Thread.sleep(Constants.DELAY);

                // retrieve the response from the given AsyncContext
                out = this.asyncContext.getResponse().getWriter();
                out.write("Phew, decomposition is setting in waiting for this to complete");
                System.out.printf("\nThread %s completed asynchronous processing", Thread.currentThread().getName());
            } catch (IOException e) {
                e.printStackTrace();
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } finally {
                if (!Objects.isNull(out)) {
                    out.flush();
                }

                this.asyncContext.complete();
            }
        }
    }
}
